@@ -1,18 +1,22 @@
-/* Copyright (C) 2006-2008 Brice Arnould.
- *  This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 3 of the License, or
- * (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
- */
+/***************************************************************************/
+/*  Copyright (C) 2006-2009 Brice Arnould.                                 */
+/*                                                                         */
+/*  This file is part of ShaKe.                                            */
+/*                                                                         */
+/*  ShaKe is free software; you can redistribute it and/or modify          */
+/*  it under the terms of the GNU General Public License as published by   */
+/*  the Free Software Foundation; either version 3 of the License, or      */
+/*  (at your option) any later version.                                    */
+/*                                                                         */
+/*  This program is distributed in the hope that it will be useful,        */
+/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
+/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
+/*  GNU General Public License for more details.                           */
+/*                                                                         */
+/*  You should have received a copy of the GNU General Public License      */
+/*  along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+/***************************************************************************/
+
 #define _GNU_SOURCE
 #include <stdlib.h>
 #include <errno.h>
@@ -24,7 +28,6 @@
 #include <dirent.h>		// scandir()
 #include <sys/stat.h>		// stat(), umask()
 #include <unistd.h>		// stat()
-#include <sys/file.h>		// flock()
 #include <stdio.h>		// printf(), tmpfile()
 #include <error.h>		// error()
 #include <limits.h>		// SSIZE_MAX
@@ -34,10 +37,6 @@
 #include "msg.h"
 
 
-
-/*  This function return a struct wich describe properties
- * of the file who's name is given.
- */
 struct accused *
 investigate (char *name, struct law *l)
 {
@@ -83,7 +82,7 @@ investigate (char *name, struct law *l)
     inode = st.st_ino;
   }
   if (!S_ISREG (a->mode) || 0 == a->size)
-    return a;
+    return a;			// a->fd is not opened or locked
   /* open() */
   if (-1 == (a->fd = open (name, O_NOATIME | O_RDWR)))
     {
@@ -91,17 +90,17 @@ investigate (char *name, struct law *l)
       goto freeall;
     }
   /* Put the lock */
-  if (l->locks && -1 == flock (a->fd, LOCK_EX | LOCK_NB))
+  if (l->locks && -1 == readlock_file (a->fd, a->name))
     {
       error (0, errno, "%s: failed to acquire a lock", a->name);
       goto freeall;
     }
-  /* the stat() will be applied only on regular files */
+  /* This stat() will be applied only on regular files */
   {
     struct stat st;
     if (-1 == fstat (a->fd, &st))
       {
-	error (0, errno, "%s: lstat() failed", name);
+	error (0, errno, "%s: fstat() failed", name);
 	goto freeall;
       }
     /* Check against race condition */
@@ -124,21 +123,23 @@ investigate (char *name, struct law *l)
     }
   if (-1 == get_testimony (a, l))
     goto freeall;
+  if (is_locked (a->fd))
+    unlock_file (a->fd);
   return a;
 freeall:
-  if (a)
-    {
-      if (a->fd != -1)
+  {
+    if (a->fd != -1)
+      {
+	if (is_locked (a->fd))
+	  unlock_file (a->fd);
 	close (a->fd);
-      free (a->name);
-      free (a);
-    }
+      }
+    free (a->name);
+    free (a);
+  }
   return NULL;
 }
 
-/*  This function free structs allocated by
- * investigate().
- */
 void
 close_case (struct accused *a, struct law *l)
 {
@@ -148,8 +149,8 @@ close_case (struct accused *a, struct law *l)
   a->name = NULL;
   if (a->fd >= 0)
     {
-      if (l->locks && -1 == flock (a->fd, LOCK_UN))
-	error (1, errno, "%s: failed to release lock", a->name);
+      if (l->locks && -1 == unlock_file (a->fd))
+	error (1, errno, "%s: failed to unlock", a->name);
       close (a->fd);
     }
   a->fd = -1;
@@ -179,7 +180,7 @@ tol_reg (struct accused *a, struct law *l)
   return tol;
 }
 
-/* Return true if the file is guilty, else false.
+/* Return true if the file is fragmented, else false.
  */
 static bool
 judge_reg (struct accused *a, struct law *l)
@@ -189,11 +190,9 @@ judge_reg (struct accused *a, struct law *l)
   double tol = tol_reg (a, l);
   if (MAX_TOL == tol)
     return false;
-  if (l->workaround && strstr (a->name, ".so"))	// TODO: investigate on ld lock
+  if (a->age < (double) l->new * tol)
     return false;
-  if (a->age < l->new * tol)
-    return false;
-  if (a->age > l->old * tol)
+  if (a->age > (double) l->old * tol)
     return true;
   if (a->fragc > l->maxfragc * tol)
     return true;
@@ -205,6 +204,8 @@ judge_reg (struct accused *a, struct law *l)
   return false;
 }
 
+/*  This function call judge on the list content
+ */
 static int
 judge_list (char *restrict * flist, struct law *restrict l)
 {
@@ -247,14 +248,14 @@ judge_list (char *restrict * flist, struct law *restrict l)
 	y->ideal = 0;
 	if (y->start)
 	  {
-	    if (x && x->end && abs (x->atime - y->atime) < MAGICTIME)
+	    if (x && x->end && labs (x->atime - y->atime) < MAGICTIME)
 	      {
-		if (z && z->start && abs (z->atime - y->atime) < MAGICTIME)
+		if (z && z->start && labs (z->atime - y->atime) < MAGICTIME)
 		  y->ideal = (x->end + z->start) / 2;
 		else
 		  y->ideal = (x->end + MAGICLEAP);
 	      }
-	    else if (z && z->start && abs (z->atime - y->atime) < MAGICTIME)
+	    else if (z && z->start && labs (z->atime - y->atime) < MAGICTIME)
 	      y->ideal = (z->start - z->blocks - MAGICLEAP);
 	  }
       }
@@ -271,6 +272,8 @@ judge_list (char *restrict * flist, struct law *restrict l)
   return res;
 }
 
+/*  This function call judge on the directory content
+ */
 static int
 judge_dir (struct accused *a, struct law *l)
 {
@@ -321,9 +324,34 @@ judge (struct accused *a, struct law *l)
     return judge_dir (a, l);
   else if (S_ISREG (a->mode) && a->size)
     {
+      /* Put the lock */
+      if (l->locks && -1 == readlock_file (a->fd, a->name))
+	{
+	  error (0, errno, "%s: failed to acquire a lock", a->name);
+	  return 0;
+	}
+      /* Check against modification */
+      {
+	struct stat st;
+	if (-1 == fstat (a->fd, &st))
+	  {
+	    error (0, errno, "%s: lstat() failed", a->name);
+	    goto freeall;
+	  }
+	if (st.st_blocks * 512 != a->size
+	    || st.st_mtime != a->mtime || st.st_mode != a->mode)
+	  {
+	    error (0, errno, "%s: concurrent access", a->name);
+	    goto freeall;
+	  }
+      }
+      /* Judge and eventually shake */
       a->guilty = judge_reg (a, l);
       if (a->guilty)
 	shake_reg (a, l);
+      /* Unlock */
+      if (is_locked (a->fd))
+	unlock_file (a->fd);
       /*  Show result of investigation, if the file is guilty or if
        * level of verbosity is greater than 2
        */
@@ -331,4 +359,7 @@ judge (struct accused *a, struct law *l)
 	show_reg (a, l);
     }
   return a->guilty;
+freeall:
+  unlock_file (a->fd);
+  return 0;
 }
